@@ -18,6 +18,7 @@ impl DyadicFraction {
             num: numerator,
             power: denominator_power,
         }
+        .canonical()
     }
 
     pub const fn zero() -> Self {
@@ -52,19 +53,54 @@ impl DyadicFraction {
         if res.num == 0 {
             res.power = 0;
         }
-        while res.num != 0 && (res.num & 1) == 0 {
+        while res.power > 0 && (res.num & 1) == 0 {
             res.power -= 1;
             res.num >>= 1;
         }
         res
     }
 
+    pub const fn round(&self, denominator_power: i8) -> Self {
+        let mut res = *self;
+        if res.num == 0 {
+            res.power = 0;
+        }
+        loop {
+            while res.power > 0 && (res.num & 1) == 0 {
+                res.power -= 1;
+                res.num >>= 1;
+            }
+            if res.power > denominator_power {
+                res.num &= !1;
+            } else {
+                break;
+            }
+        }
+        res
+    }
+
+    pub fn floor(&self) -> i32 {
+        let val = self.canonical();
+        let shift = val.power.abs();
+        if val.power <= 0 {
+            saturating_shl(val.num, shift)
+        } else {
+            saturating_shr(val.num, shift)
+        }
+    }
+
+    pub fn div_by_two(&self) -> Self {
+        let mut res = *self;
+        res.power = res.power.saturating_add(1);
+        res
+    }
+
     pub fn mul_add(self, a: Self, b: Self) -> Self {
         self * a + b
     }
 
     pub fn scale(self, a: impl Into<Self>) -> i32 {
-        self.mul(a.into()).into()
+        self.canonical().mul(a.into()).floor()
     }
 
     pub fn pow(self, n: u8) -> Self {
@@ -78,14 +114,6 @@ impl DyadicFraction {
         res
     }
 
-    pub fn numerator(&self) -> i32 {
-        self.num
-    }
-
-    pub fn denominator_power(&self) -> i8 {
-        self.power
-    }
-
     pub fn max(lhs: Self, rhs: Self) -> Self {
         if lhs > rhs {
             lhs
@@ -102,44 +130,43 @@ impl DyadicFraction {
         }
     }
 
-    fn cross(self, other: Self) -> (i32, i32, i8) {
+    pub fn numerator(&self) -> i32 {
+        self.num
+    }
+
+    pub fn denominator_power(&self) -> i8 {
+        self.power
+    }
+
+    fn saturating_cross(self, other: Self) -> (i32, i32, i8) {
         let (min_power, max_power) = if self.power > other.power {
             (other.power, self.power)
         } else {
             (self.power, other.power)
         };
-        let power = max_power;
-        let fst = if (other.power - min_power) < 32 {
-            self.num.shl(other.power - min_power)
-        } else {
-            self.num.signum() * i32::MAX
-        };
-        let snd = if (self.power - min_power) < 32 {
-            other.num.shl(self.power - min_power)
-        } else {
-            other.num.signum() * i32::MAX
-        };
-        (fst, snd, power)
+        (
+            saturating_shl(self.num, other.power - min_power),
+            saturating_shl(other.num, self.power - min_power),
+            max_power,
+        )
     }
 }
 
-impl Into<i32> for DyadicFraction {
-    fn into(self) -> i32 {
-        let val = self.canonical();
-        let shift = val.power.abs();
-        if shift > 32 {
-            if val.power.is_negative() {
-                val.num.signum() * i32::MAX
-            } else {
-                0
-            }
-        } else {
-            if val.power.is_negative() {
-                val.num.shl(shift)
-            } else {
-                val.num.shr(shift)
-            }
-        }
+fn saturating_shl(num: i32, rhs: i8) -> i32 {
+    if rhs < 32 {
+        num.shl(rhs)
+    } else if num.is_positive() {
+        i32::MAX
+    } else {
+        i32::MIN
+    }
+}
+
+fn saturating_shr(num: i32, rhs: i8) -> i32 {
+    if rhs < 32 {
+        num.shr(rhs)
+    } else {
+        0
     }
 }
 
@@ -149,6 +176,12 @@ impl From<i32> for DyadicFraction {
     }
 }
 
+impl From<isize> for DyadicFraction {
+    fn from(num: isize) -> Self {
+        Self::new(num as _, 0)
+    }
+}
+
 impl From<i16> for DyadicFraction {
     fn from(num: i16) -> Self {
         Self::new(num as _, 0)
@@ -177,7 +210,7 @@ impl Add for DyadicFraction {
     type Output = Self;
 
     fn add(self, other: Self) -> Self {
-        let (fst, snd, power) = self.cross(other);
+        let (fst, snd, power) = self.saturating_cross(other);
         Self::new(fst.saturating_add(snd), power)
     }
 }
@@ -192,7 +225,7 @@ impl Sub for DyadicFraction {
     type Output = Self;
 
     fn sub(self, other: Self) -> Self {
-        let (fst, snd, power) = self.cross(other);
+        let (fst, snd, power) = self.saturating_cross(other);
         Self::new(fst.saturating_sub(snd), power)
     }
 }
@@ -211,6 +244,7 @@ impl Mul for DyadicFraction {
             self.num.saturating_mul(other.num),
             self.power.saturating_add(other.power),
         )
+        .canonical()
     }
 }
 
@@ -246,37 +280,14 @@ impl PartialOrd for DyadicFraction {
 
 impl Ord for DyadicFraction {
     fn cmp(&self, other: &Self) -> Ordering {
-        let shift = other.power.abs();
-        let lhs = if other.power.is_negative() {
-            if shift < 32 {
-                self.num.shr(shift)
-            } else {
-                0
-            }
+        let delta = (*self - *other).numerator().signum();
+        if delta == 0 {
+            Ordering::Equal
+        } else if delta.is_positive() {
+            Ordering::Greater
         } else {
-            if shift < 32 {
-                self.num.shl(shift)
-            } else {
-                self.num.signum() * i32::MAX
-            }
-        };
-
-        let shift = self.power.abs();
-        let rhs = if self.power.is_negative() {
-            if shift < 32 {
-                other.num.shr(shift)
-            } else {
-                0
-            }
-        } else {
-            if shift < 32 {
-                other.num.shl(shift)
-            } else {
-                other.num.signum() * i32::MAX
-            }
-        };
-
-        lhs.cmp(&rhs)
+            Ordering::Less
+        }
     }
 }
 
@@ -284,20 +295,67 @@ impl fmt::Display for DyadicFraction {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let val = self.canonical();
         let shift = val.power.abs();
-        if val.power == 0 || val.power.is_negative() {
-            let val = if shift < 32 {
-                val.num.shl(shift)
-            } else {
-                val.num.signum() * i32::MAX
-            };
-            write!(f, "{}", val)
+        if val.power <= 0 {
+            write!(f, "{}", saturating_shl(val.num, shift))
         } else {
-            let den = if shift < 32 {
-                1.shl(shift)
-            } else {
-                i32::MAX
-            };
-            write!(f, "{}/{}", val.num, den)
+            write!(f, "{}/{}", val.num, saturating_shl(1, shift))
         }
     }
 }
+
+pub mod consts {
+    use super::*;
+
+    /// The full circle constant (τ)
+    ///
+    /// Equal to 2π.
+    pub const TAU: DyadicFraction = DyadicFraction::new(3217, 9);
+
+    /// Archimedes' constant (π)
+    pub const PI: DyadicFraction = DyadicFraction::new(3217, 10);
+
+    /// π/2
+    pub const FRAC_PI_2: DyadicFraction = DyadicFraction::new(3217, 11);
+
+    /// π/4
+    pub const FRAC_PI_4: DyadicFraction = DyadicFraction::new(3217, 12);
+
+    /// π/3
+    pub const FRAC_PI_3: DyadicFraction = DyadicFraction::new(68629, 16);
+
+    /// 1/π
+    pub const FRAC_1_PI: DyadicFraction = DyadicFraction::new(20861, 16);
+
+    /// 2/π
+    pub const FRAC_2_PI: DyadicFraction = DyadicFraction::new(20861, 15);
+
+    /// 2/sqrt(π)
+    pub const FRAC_2_SQRT_PI: DyadicFraction = DyadicFraction::new(73949, 16);
+
+    /// sqrt(2)
+    pub const SQRT_2: DyadicFraction = DyadicFraction::new(46341, 15);
+
+    /// 1/sqrt(2)
+    pub const FRAC_1_SQRT_2: DyadicFraction = DyadicFraction::new(46341, 16);
+
+    /// Euler's number (e)
+    pub const E: DyadicFraction = DyadicFraction::new(178145, 16);
+
+    /// log<sub>2</sub>(10)
+    pub const LOG2_10: DyadicFraction = DyadicFraction::new(108853, 15);
+
+    /// log<sub>2</sub>(e)
+    pub const LOG2_E: DyadicFraction = DyadicFraction::new(23637, 14);
+
+    /// log<sub>10</sub>(2)
+    pub const LOG10_2: DyadicFraction = DyadicFraction::new(1233, 12);
+
+    /// log<sub>10</sub>(e)
+    pub const LOG10_E: DyadicFraction = DyadicFraction::new(14231, 15);
+
+    /// ln(2)
+    pub const LN_2: DyadicFraction = DyadicFraction::new(22713, 15);
+
+    /// ln(10)
+    pub const LN_10: DyadicFraction = DyadicFraction::new(75451, 15);
+}
